/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; coding: utf-8 -*- 
 *
 * Copyright © 2007 Björn Lindqvist <bjourne@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
#include "image_view_drawer.h"
#include "utils.h"

static gboolean
draw_settings_equals (DrawSettings *d1,
                      DrawSettings *d2)
{
    if (d1->zoom == d2->zoom &&
        d1->pos.x == d2->pos.x &&
        d1->pos.y == d2->pos.y &&
        d1->interp == d2->interp &&
        gdk_rectangle_eq (d1->rect, d2->rect) &&
        d1->pixbuf == d2->pixbuf &&
        d1->check_color1 == d2->check_color1 &&
        d1->check_color2 == d2->check_color2)
    {
        return TRUE;
    }
    return FALSE;
}

/**
 * draw_settings_get_flags:
 * @old: The draw settings used in the last draw operation.
 * @new: The draw settings used in the current draw operation.
 * @last_pixbuf: The pixbuf to be drawn.
 *
 * Returns which actions that needs to be carried out in this draw.
 **/
DrawFlags
draw_settings_get_flags (DrawSettings *old,
                         DrawSettings *new,
                         GdkPixbuf    *last_pixbuf)
{
    GdkColorspace new_cs = gdk_pixbuf_get_colorspace (new->pixbuf);
    GdkColorspace old_cs = gdk_pixbuf_get_colorspace (old->pixbuf);
    int new_bps = gdk_pixbuf_get_bits_per_sample (new->pixbuf);
    int old_bps = gdk_pixbuf_get_bits_per_sample (old->pixbuf);
    if (new_cs != old_cs ||
        new_bps != old_bps ||
        new->rect.width > gdk_pixbuf_get_width (last_pixbuf) ||
        new->rect.height > gdk_pixbuf_get_height (last_pixbuf))
        return DRAW_FLAGS_ALLOCATE | DRAW_FLAGS_SCALE;
    if (!draw_settings_equals (new, old))
        return DRAW_FLAGS_SCALE;
    return 0;
}

ImageViewDrawer *
image_view_drawer_new ()
{
    ImageViewDrawer *drawer = g_new0 (ImageViewDrawer, 1);
    drawer->last_pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 1, 1);
    drawer->check_size = 16;
    drawer->old = (DrawSettings){0, {0, 0}, 0, drawer->last_pixbuf,
                                 {0, 0, 0, 0}, 0, 0};
    return drawer;
}

void
image_view_drawer_free (ImageViewDrawer *drawer)
{
    g_object_unref (drawer->last_pixbuf);
    g_free (drawer);
}

/**
 * image_view_drawer_force_scale:
 *
 * This method forces the drawer to perform a %DRAW_FLAGS_SCALE
 * operation the next time image_view_drawer_draw() is called.
 *
 * ImageViewDrawer tries to minimize the number of scale operations
 * needed by caching the last drawn pixbuf. It would be inefficient to
 * check the individual pixels inside the pixbuf so it assumes that if
 * the memory address of the pixbuf has not changed, then the cache is
 * good to use.
 *
 * Use this method to override that assumption.
 **/
void
image_view_drawer_force_scale (ImageViewDrawer *drawer)
{
    /* Set the cached zoom to a bogus value, to force a
       DRAW_FLAGS_SCALE but not a DRAW_FLAGS_ALLOCATE. */
    drawer->old.zoom = -1234.0;
}

/**
 * image_view_drawer_draw:
 * @drawer: An #ImageViewDrawer.
 * @ds: The #DrawSettings to use for this redraw.
 **/
void
image_view_drawer_draw (ImageViewDrawer *drawer,
                        DrawSettings    *ds,
                        GdkDrawable     *drawable)
{
    DrawFlags flags = draw_settings_get_flags (&drawer->old, ds,
                                               drawer->last_pixbuf);
    if (flags & DRAW_FLAGS_ALLOCATE)
    {
        g_object_unref (drawer->last_pixbuf);
        GdkColorspace cs = gdk_pixbuf_get_colorspace (ds->pixbuf);
        int bps = gdk_pixbuf_get_bits_per_sample (ds->pixbuf);
        drawer->last_pixbuf = gdk_pixbuf_new (cs, FALSE, bps,
                                              ds->rect.width,
                                              ds->rect.height);
    }
    if (flags & DRAW_FLAGS_SCALE)
    {
        if (gdk_pixbuf_get_has_alpha (ds->pixbuf))
            gdk_pixbuf_composite_color (ds->pixbuf,
                                        drawer->last_pixbuf,
                                        0, 0,
                                        ds->rect.width, ds->rect.height,
                                        (double) -ds->pos.x, (double) -ds->pos.y,
                                        ds->zoom, ds->zoom,
                                        ds->interp,
                                        255,
                                        ds->pos.x, ds->pos.y,
                                        drawer->check_size,
                                        ds->check_color1,
                                        ds->check_color2);
            
        else
        {
            gdk_pixbuf_scale (ds->pixbuf,
                              drawer->last_pixbuf,
                              0, 0,
                              ds->rect.width, ds->rect.height,
                              -ds->pos.x, -ds->pos.y,
                              ds->zoom, ds->zoom,
                              ds->interp);
        }
    }
    gdk_draw_pixbuf (drawable,
                     NULL,
                     drawer->last_pixbuf,
                     0, 0,
                     ds->rect.x, ds->rect.y,
                     ds->rect.width, ds->rect.height,
                     GDK_RGB_DITHER_MAX,
                     ds->rect.x, ds->rect.y);
    drawer->old = *ds;
}

