/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; coding: utf-8 -*- */
/**
 * This file contains tests that validates that the ImageViewDrawer
 * class works correctly.
 **/
#include <assert.h>
#include <src/gtkimageview.h>

/**
 * test_only_scale_op_on_new_identical_pixbuf:
 *
 * The objective of this test is to verify that the cached pixbuf that
 * the ImageViewDrawer holds is not discarded when it receives a new
 * pixbuf with the exact same bit depth and colorspace.
 **/
static void
test_only_scale_op_on_new_identical_pixbuf ()
{
    printf ("test_only_scale_op_on_new_identical_pixbuf\n");

    GdkPixbuf *pixbuf1 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 9, 9);
    GdkPixbuf *pixbuf2 = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 9, 9);
    GdkPixbuf *last_pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 50, 50);
    
    Pos p = {0, 0};
    int interp = GDK_INTERP_BILINEAR;
    GdkRectangle area = {0, 0, 10, 10};
    
    DrawSettings parms1 = {1, p, interp, pixbuf1, area};
    DrawSettings parms2 = {1, p, interp, pixbuf2, area};

    DrawFlags flags = draw_settings_get_flags (&parms1, &parms2, last_pb);
    assert (flags == DRAW_FLAGS_SCALE);

    g_object_unref (pixbuf1);
    g_object_unref (pixbuf2);
    g_object_unref (last_pb);
}

/**
 * test_cache_is_used_on_equal_settings:
 *
 * The objective of this test is to verify that the cache is used when
 * the new draw settings are identical to the old ones if the area is
 * smaller than the cached pixbuf.
 **/
static void
test_cache_is_used_on_equal_settings ()
{
    printf ("test_cache_is_used_on_equal_settings\n");
    Pos p = {0, 0};
    int interp = GDK_INTERP_BILINEAR;
    GdkRectangle area = {0, 0, 10, 10};
    GdkPixbuf *pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 30, 30);
    DrawSettings parms1 = {1, p, interp, pb, area};
    DrawSettings parms2 = {1, p, interp, pb, area};

    DrawFlags flags = draw_settings_get_flags (&parms1, &parms2, pb);
    assert (flags == 0);

    g_object_unref (pb);
}

/**
 * test_cache_not_used_if_to_small:
 *
 * The objective of this test is to verify that the cached pixbuf is
 * not used if its size is smaller than the area to be drawn.
 **/
static void
test_cache_not_used_if_to_small ()
{
    printf ("test_cache_not_used_if_to_small\n");
    Pos p = {0, 0};
    int interp = GDK_INTERP_BILINEAR;
    GdkRectangle area = {0, 0, 20, 20};
    
    GdkPixbuf *draw_pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 30, 30);
    DrawSettings parms1 = {1, p, interp, draw_pb, area};
    DrawSettings parms2 = {1, p, interp, draw_pb, area};

    GdkPixbuf *last_pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 10, 10);
    DrawFlags flags = draw_settings_get_flags (&parms1, &parms2, last_pb);
    assert (flags == (DRAW_FLAGS_ALLOCATE | DRAW_FLAGS_SCALE));
    g_object_unref (draw_pb);
    g_object_unref (last_pb);
}


/**
 * test_scale_needed_if_rect_size_not_equal:
 *
 * The objective of this test is to verify that a scale operation is
 * needed if the size of the rectangle changes.
 **/
static void
test_scale_needed_if_rect_size_not_equal ()
{
    printf ("test_scale_needed_if_rect_size_not_equal\n");
    Pos p = {0, 0};
    int interp = GDK_INTERP_BILINEAR;
    GdkPixbuf *pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 9, 9);
    DrawSettings parms1 = {1, p, interp, pb, (GdkRectangle){0, 0, 20, 20}};
    DrawSettings parms2 = {1, p, interp, pb, (GdkRectangle){0, 0, 10, 10}};

    DrawFlags flags = draw_settings_get_flags (&parms1, &parms2, pb);
    assert (flags & DRAW_FLAGS_SCALE);

    g_object_unref (pb);
}

/**
 * test_default_draw_settings:
 *
 * The objective of this test is to verify that the default draw
 * settings in the image drawer results in the correct draw flags.
 **/
static void
test_default_draw_settings ()
{
    printf ("test_default_draw_settings\n");
    ImageViewDrawer *drawer = image_view_drawer_new ();
    Pos p = {0, 0};
    int interp = GDK_INTERP_BILINEAR;
    GdkPixbuf *pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 9, 9);
    DrawSettings parms = {1, p, interp, pb, (GdkRectangle){0, 0, 20, 20}};
    DrawFlags flags = draw_settings_get_flags (&drawer->old, &parms,
                                               drawer->last_pixbuf);
    assert (flags == (DRAW_FLAGS_ALLOCATE | DRAW_FLAGS_SCALE));
    
    image_view_drawer_free (drawer);
    g_object_unref (pb);
}

/**
 * test_force_scale_operation:
 *
 * The objective of this test is to verify that forcing the
 * ImageViewDrawer to apply a scale operation works as expected.
 **/
static void
test_force_scale_operation ()
{
    printf ("test_force_scale_operation\n");
    ImageViewDrawer *drawer = image_view_drawer_new ();
    
    Pos p = {0, 0};
    GdkInterpType interp = GDK_INTERP_BILINEAR;
    GdkPixbuf *pb = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 9, 9);
    DrawSettings parms = {1, p, interp, pb, (GdkRectangle){0, 0, 5, 5}};
    drawer->old = parms;


    /* DrawSettings are identical, so no allocation or scale necessary. */
    assert (draw_settings_get_flags (&drawer->old, &parms, pb) == 0);

    /* Force a scale operation in the next draw. */
    image_view_drawer_force_scale (drawer);

    DrawFlags flags = draw_settings_get_flags (&drawer->old, &parms, pb);
    assert (flags == DRAW_FLAGS_SCALE);
    
    image_view_drawer_free (drawer);
    g_object_unref (pb);
}

int
main(int argc, char *argv[])
{
    gtk_init (&argc, &argv);
    test_only_scale_op_on_new_identical_pixbuf ();
    test_cache_is_used_on_equal_settings ();
    test_cache_not_used_if_to_small ();
    test_scale_needed_if_rect_size_not_equal ();
    test_default_draw_settings ();
    test_force_scale_operation ();
    printf ("6 tests passed.\n");
}
